from __future__ import annotations

# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

EXT_M3U = "#EXTM3U"
EXT_X_TARGET_DURATION = "#EXT-X-TARGETDURATION"
EXT_X_MEDIA_SEQUENCE = "#EXT-X-MEDIA-SEQUENCE"
EXT_X_DISCONTINUITY_SEQUENCE = "#EXT-X-DISCONTINUITY-SEQUENCE"
EXT_X_PROGRAM_DATE_TIME = "#EXT-X-PROGRAM-DATE-TIME"
EXT_X_MEDIA = "#EXT-X-MEDIA"
EXT_X_PLAYLIST_TYPE = "#EXT-X-PLAYLIST-TYPE"
EXT_X_KEY = "#EXT-X-KEY"
EXT_X_STREAM_INF = "#EXT-X-STREAM-INF"
EXT_X_VERSION = "#EXT-X-VERSION"
EXT_X_ALLOW_CACHE = "#EXT-X-ALLOW-CACHE"
EXT_X_END_LIST = "#EXT-X-ENDLIST"
EXT_INF = "#EXTINF"
EXT_I_FRAMES_ONLY = "#EXT-X-I-FRAMES-ONLY"
EXT_X_BIT_RATE = "#EXT-X-BITRATE"
EXT_X_BYTE_RANGE = "#EXT-X-BYTERANGE"
EXT_X_I_FRAME_STREAM_INF = "#EXT-X-I-FRAME-STREAM-INF"
EXT_X_DISCONTINUITY = "#EXT-X-DISCONTINUITY"
EXT_X_CUE_OUT = "#EXT-X-CUE-OUT"
EXT_X_CUE_OUT_CONT = "#EXT-X-CUE-OUT-CONT"
EXT_X_CUE_IN = "#EXT-X-CUE-IN"
EXT_X_CUE_SPAN = "#EXT-X-CUE-SPAN"
EXT_X_SCTE35 = "#EXT-OATCLS-SCTE35"
EXT_IS_INDEPENDENT_SEGMENTS = "#EXT-X-INDEPENDENT-SEGMENTS"
EXT_X_MAP = "#EXT-X-MAP"
EXT_X_START = "#EXT-X-START"
EXT_X_SERVER_CONTROL = "#EXT-X-SERVER-CONTROL"
EXT_X_PART_INF = "#EXT-X-PART-INF"
EXT_X_PART = "#EXT-X-PART"
EXT_X_RENDITION_REPORT = "#EXT-X-RENDITION-REPORT"
EXT_X_SKIP = "#EXT-X-SKIP"
EXT_X_SESSION_DATA = "#EXT-X-SESSION-DATA"
EXT_X_SESSION_KEY = "#EXT-X-SESSION-KEY"
EXT_X_PRELOAD_HINT = "#EXT-X-PRELOAD-HINT"
EXT_X_DATERANGE = "#EXT-X-DATERANGE"
EXT_X_GAP = "#EXT-X-GAP"
EXT_X_CONTENT_STEERING = "#EXT-X-CONTENT-STEERING"



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING, PYLAV_BUNDLED_NODES_SETTINGS
from pylav.logging import getLogger
from pylav.storage.database.tables.nodes import NodeRow
from pylav.storage.models.node.mocked import NodeMock
from pylav.storage.models.node.real import Node

if TYPE_CHECKING:
    from pylav.core.client import Client
LOGGER = getLogger("PyLav.Database.Controller.Node")


class NodeController:
    __slots__ = ("_client", "currently_in_db")

    def __init__(self, client: Client) -> None:
        self._client = client
        self.currently_in_db = {
            self._client.bot.user.id,
        }

    @property
    def client(self) -> Client:
        return self._client

    def bundled_node_config(self) -> Node:
        return Node(id=self._client.bot.user.id)

    def get_node_config(self, node_id: int) -> Node:
        if node_id not in BUNDLED_NODES_IDS_HOST_MAPPING:
            node = Node(id=node_id)
        else:
            node = NodeMock(id=node_id, data=PYLAV_BUNDLED_NODES_SETTINGS[BUNDLED_NODES_IDS_HOST_MAPPING[node_id]])
        self.currently_in_db.add(node.id)
        return node

    async def get_all_unmanaged_nodes(self, dedupe: bool = True) -> list[Node]:

        model_list = [
            Node(**node)
            for node in await NodeRow.select(NodeRow.id)
            .where(
                (NodeRow.managed == False)  # noqa: E712
                & (NodeRow.id.not_in(list(BUNDLED_NODES_IDS_HOST_MAPPING.keys())))
            )
            .output(nested=True, load_json=True)
        ]
        new_model_list = list(set(model_list)) if dedupe else model_list
        for n in new_model_list:
            self.currently_in_db.add(n.id)
        return new_model_list

    async def get_all_nodes(self) -> list[Node]:
        model_list = await self.get_all_unmanaged_nodes(dedupe=True)
        if mn := self.bundled_node_config():
            model_list.append(mn)
        return model_list

    async def get_bundled_node_config(self) -> Node | None:
        response = (
            await NodeRow.select(NodeRow.id)
            .where((NodeRow.id == self._client.bot.user.id) & (NodeRow.managed is True))
            .first()
        )
        return Node(**response) if response else None

    async def update_node(
        self,
        host: str,
        port: int,
        password: str,
        unique_identifier: int,
        resume_timeout: int = 60,
        name: str = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        extras: dict = None,
        yaml: dict = None,
        disabled_sources: list[str] = None,
    ) -> Node:

        """
        Add a new node to the database.

        Parameters
        ----------
        host: str
            The host of the node.
        port: int
            The port of the node.
        password: str
            The password of the node.
        unique_identifier: int
            The unique identifier of the node.
        name: str
            The name of the node.
        resume_timeout: int
            The resume timeout of the node.
        reconnect_attempts: int
            The reconnect attempts of the node.
        ssl: bool
            Whether the node is using ssl.
        search_only: bool
            Whether the node is search only.
        managed: bool
            Whether the node is managed.
        extras: dict
            The extras of the node.
        disabled_sources: list[str]
            The disabled sources of the node.
        yaml: dict
            The yaml of the node.
        """

        node = Node(id=unique_identifier)
        self.currently_in_db.add(node.id)
        await node.bulk_update(
            host=host,
            port=port,
            password=password,
            name=name,
            resume_timeout=resume_timeout,
            reconnect_attempts=reconnect_attempts,
            ssl=ssl,
            search_only=search_only,
            managed=managed,
            extras=extras,
            disabled_sources=disabled_sources,
            yaml=yaml,
        )
        return node

    async def delete(self, node_id: int) -> None:
        if node_id == self._client.bot.user.id:
            raise ValueError("Cannot delete bundled node")
        await self.get_node_config(node_id=node_id).delete()

    @staticmethod
    async def count() -> int:
        """Return the number of unbundled nodes in the database."""
        return await NodeRow.count().where(NodeRow.id.not_in(set(BUNDLED_NODES_IDS_HOST_MAPPING.keys())))



from __future__ import annotations

from pylav.exceptions.base import PyLavException
from pylav.nodes.api.responses.errors import LavalinkError


class HTTPException(PyLavException):
    """Base exception for HTTP request errors"""

    def __init__(self, response: LavalinkError):
        self.response = response


class UnauthorizedException(HTTPException):
    """Raised when a REST request fails due to an incorrect password"""



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

import asyncstdlib
import discord
from redbot.core.i18n import Translator
from redbot.vendored.discord.ext import menus

from pylav.extension.red.ui.selectors.options.queue import QueueTrackOption, SearchTrackOption
from pylav.logging import getLogger
from pylav.players.tracks.obj import Track
from pylav.type_hints.bot import DISCORD_COG_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.queue import QueueMenu, QueuePickerMenu

LOGGER = getLogger("PyLav.ext.red.ui.sources.queue")

_ = Translator("PyLav", Path(__file__))


class SearchPickerSource(menus.ListPageSource):
    entries: list[Track]

    def __init__(self, entries: list[Track], cog: DISCORD_COG_TYPE, per_page: int = 10):
        super().__init__(entries=entries, per_page=per_page)
        self.per_page = 25
        self.select_options: list[SearchTrackOption] = []
        self.cog = cog
        self.select_mapping: dict[str, Track] = {}

    async def get_page(self, page_number):
        if page_number > self.get_max_pages():
            page_number = 0
        base = page_number * self.per_page
        self.select_options.clear()
        self.select_mapping.clear()
        async for i, track in asyncstdlib.enumerate(
            asyncstdlib.iter(self.entries[base : base + self.per_page]), start=base
        ):  # noqa: E203
            self.select_options.append(await SearchTrackOption.from_track(track=track, index=i))
            self.select_mapping[track.id] = track
        return []

    async def format_page(self, menu: QueueMenu, entries: list[Track]) -> str:
        return ""

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence"""
        return self._max_pages or 1


class QueueSource(menus.ListPageSource):
    def __init__(self, guild_id: int, cog: DISCORD_COG_TYPE, history: bool = False):  # noqa
        self.cog = cog
        self.current_player = None
        self.per_page = 10
        self.guild_id = guild_id
        self.history = history

    @property
    def entries(self) -> Iterable[Track]:
        if player := self.cog.pylav.get_player(self.guild_id):
            return player.history.raw_queue if self.history else player.queue.raw_queue
        else:
            return []

    def is_paginating(self) -> bool:
        return True

    async def get_page(self, page_number: int) -> list[Track]:
        base = page_number * self.per_page
        return await asyncstdlib.list(asyncstdlib.islice(self.entries, base, base + self.per_page))

    def get_max_pages(self) -> int:
        player = self.cog.pylav.get_player(self.guild_id)
        if not player:
            return 1
        pages, left_over = divmod(player.history.size() if self.history else player.queue.size(), self.per_page)

        if left_over:
            pages += 1
        return pages or 1

    def get_starting_index_and_page_number(self, menu: QueueMenu) -> tuple[int, int]:
        page_num = menu.current_page
        start = page_num * self.per_page
        return start, page_num

    async def format_page(self, menu: QueueMenu, tracks: list[Track]) -> discord.Embed:
        if not (player := self.cog.pylav.get_player(menu.ctx.guild.id)):
            return await self.cog.pylav.construct_embed(
                description=_("No active player found in server"), messageable=menu.ctx
            )
        self.current_player = player
        return (
            await player.get_queue_page(
                page_index=menu.current_page,
                per_page=self.per_page,
                total_pages=self.get_max_pages(),
                embed=True,
                messageable=menu.ctx,
                history=self.history,
            )
            if player.current and (player.history.size() if self.history else True)
            else await self.cog.pylav.construct_embed(
                description=_("There's nothing in recently played")
                if self.history
                else _("There's nothing currently being played"),
                messageable=menu.ctx,
            )
        )


class QueuePickerSource(QueueSource):
    def __init__(self, guild_id: int, cog: DISCORD_COG_TYPE):
        super().__init__(guild_id, cog=cog)
        self.per_page = 25
        self.select_options: list[QueueTrackOption] = []
        self.select_mapping: dict[str, Track] = {}
        self.cog = cog

    async def get_page(self, page_number):
        if page_number > self.get_max_pages():
            page_number = 0
        base = page_number * self.per_page
        self.select_options.clear()
        self.select_mapping.clear()
        async for i, track in asyncstdlib.enumerate(
            asyncstdlib.islice(self.entries, base, base + self.per_page), start=base
        ):
            self.select_options.append(await QueueTrackOption.from_track(track=track, index=i))
            self.select_mapping[track.id] = track
        return []

    async def format_page(self, menu: QueuePickerMenu, tracks: list[Track]) -> discord.Embed:
        if not (player := self.cog.pylav.get_player(menu.ctx.guild.id)):
            return await self.cog.pylav.construct_embed(
                description=_("No active player found in server"), messageable=menu.ctx
            )
        self.current_player = player
        return (
            await player.get_queue_page(
                page_index=menu.current_page,
                per_page=self.per_page,
                total_pages=self.get_max_pages(),
                embed=True,
                messageable=menu.ctx,
            )
            if player.current
            else await self.cog.pylav.construct_embed(
                description=_("There's nothing currently being played"),
                messageable=menu.ctx,
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.extension.m3u.base import load, loads

if TYPE_CHECKING:
    from pylav.core.client import Client


class M3UParser:
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    load = load
    loads = loads



from __future__ import annotations

import dataclasses
import typing

from packaging.version import Version as _Version
from packaging.version import parse

from pylav.constants.regex import GIT_SHA1, SEMANTIC_VERSIONING
from pylav.constants.versions import API_DEVELOPMENT_VERSION


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Version:
    semver: str
    major: int
    minor: int
    patch: int
    preRelease: str | None = None

    def __post_init__(self) -> None:
        if not SEMANTIC_VERSIONING.match(self.semver):
            sha1 = GIT_SHA1.search(self.semver)
            sha1 = sha1.group("sha1") if sha1 else "unknown"
            version = _Version(f"{API_DEVELOPMENT_VERSION}+{sha1}")
        else:
            version = typing.cast(_Version, parse(self.semver))
        object.__setattr__(self, "semver", version)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Git:
    branch: str
    commit: str
    commitTime: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Plugin:
    name: str
    version: str



from __future__ import annotations

import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LowPass:
    smoothing: Annotated[float | None, ValueRange(min=1.0, max=float("inf"))] = None

    def to_dict(self) -> dict[str, float]:
        return {} if self.smoothing is None else {"smoothing": self.smoothing}



from __future__ import annotations

import dataclasses
from typing import Literal, TypeAlias, Union  # noqa

from pylav.nodes.api.responses.filters import Filters
from pylav.nodes.api.responses.misc import Git, Plugin, Version
from pylav.nodes.api.responses.playlists import Info
from pylav.nodes.api.responses.shared import PlaylistPluginInfo
from pylav.nodes.api.responses.track import Track
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LoadException:
    severity: Literal["COMMON", "SUSPICIOUS", "FAULT"]
    message: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkException(LoadException):
    cause: str | None = None  # This is only optional so that inheritance in python works


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class BaseTrackResponse:
    playlistInfo: None | Info | dict = None
    pluginInfo: None | PlaylistPluginInfo | dict = None
    exception: None | LoadException | dict = None
    tracks: list[Track] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.pluginInfo, dict):
            object.__setattr__(self, "pluginInfo", PlaylistPluginInfo(kwargs=self.pluginInfo))
        elif self.pluginInfo is None:
            object.__setattr__(self, "pluginInfo", PlaylistPluginInfo(kwargs=None))
        if isinstance(self.exception, dict):
            object.__setattr__(self, "exception", LoadException(**self.exception))
        if isinstance(self.playlistInfo, dict):
            object.__setattr__(self, "playlistInfo", Info(**self.playlistInfo))
        temp = []
        for s in self.tracks:
            if isinstance(s, Track) or (isinstance(s, dict) and (s := Track(**s))):
                temp.append(s)

        object.__setattr__(self, "tracks", temp)


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackLoaded(BaseTrackResponse):  # noqa
    loadType: Literal["TRACK_LOADED"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlaylistLoaded(BaseTrackResponse):  # noqa
    loadType: Literal["PLAYLIST_LOADED"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class SearchResult(BaseTrackResponse):  # noqa
    loadType: Literal["SEARCH_RESULT"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class NoMatches(BaseTrackResponse):  # noqa
    loadType: Literal["NO_MATCHES"]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LoadFailed(BaseTrackResponse):  # noqa
    loadType: Literal["LOAD_FAILED"]


LoadTrackResponses: TypeAlias = Union[
    TrackLoaded,
    PlaylistLoaded,
    NoMatches,
    LoadFailed,
    SearchResult,
]


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkInfo:
    version: Version
    buildTime: int
    git: Git
    jvm: str
    lavaplayer: str
    sourceManagers: list[str]
    filters: list[str]
    plugins: list[Plugin]

    def __post_init__(self):
        temp = []
        for s in self.plugins:
            if isinstance(s, Plugin) or (isinstance(s, dict) and (s := Plugin(**s))):
                temp.append(s)
        object.__setattr__(self, "plugins", temp)


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class VoiceState:
    token: str
    endpoint: str
    sessionId: str
    connected: bool | None = None
    ping: int | None = -1

    def to_dict(self) -> JSON_DICT_TYPE:
        return {
            "token": self.token,
            "endpoint": self.endpoint,
            "sessionId": self.sessionId,
            "connected": self.connected,
            "ping": self.ping,
        }

    def __repr__(self) -> str:
        return (
            f"<VoiceStateObject(token={'OBFUSCATED' if self.token else None} "
            f"endpoint={self.endpoint} "
            f"sessionId={self.sessionId} "
            f"connected={self.connected} "
            f"ping={self.ping})"
        )


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class LavalinkPlayer:
    guildId: str
    volume: int
    paused: bool
    voice: VoiceState | dict
    filters: Filters | dict
    track: Track | dict | None = None

    def __post_init__(self) -> None:
        if isinstance(self.voice, dict):
            object.__setattr__(self, "voice", VoiceState(**self.voice))
        if isinstance(self.filters, dict):
            object.__setattr__(self, "filters", Filters(**self.filters))
        if isinstance(self.track, dict):
            object.__setattr__(self, "track", Track(**self.track))

    def to_dict(
        self,
    ) -> JSON_DICT_TYPE:
        assert (
            isinstance(self.voice, VoiceState)
            and isinstance(self.filters, Filters)
            and isinstance(self.track, (Track, type(None)))
        )
        return {
            "guildId": self.guildId,
            "volume": self.volume,
            "paused": self.paused,
            "voice": self.voice.to_dict(),
            "filters": self.filters.to_dict(),
            "track": self.track.to_dict() if self.track else None,
        }



from __future__ import annotations

import dataclasses


@dataclasses.dataclass(repr=True, frozen=True, slots=True)
class ValueRange:
    min: float
    max: float

    def validate_value(self, x: float) -> None:
        if not (self.min <= x <= self.max):
            raise ValueError(f"{x} must be in range({self.min}, {self.max})")



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.versions import VERSION_0_0_0, VERSION_0_3_2_0
from pylav.storage.migrations.logging import LOGGER

if TYPE_CHECKING:
    from pylav.core.client import Client


async def migration_v_0_3_2_0(client: Client, current_version: Version) -> None:
    if current_version >= VERSION_0_3_2_0 or current_version == VERSION_0_0_0:
        return

    LOGGER.info("Running %s migration", VERSION_0_3_2_0)
    config = client.node_db_manager.bundled_node_config()
    yaml_data = await config.fetch_yaml()
    yaml_data["lavalink"]["server"]["opusEncodingQuality"] = 10
    yaml_data["lavalink"]["server"]["resamplingQuality"] = "LOW"
    yaml_data["lavalink"]["server"]["useSeekGhosting"] = True
    await config.update_yaml(yaml_data)
    await client.lib_db_manager.update_bot_dv_version(VERSION_0_3_2_0)



from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import TYPE_CHECKING

from iso8601 import iso8601

from pylav.players.query.obj import Query

if TYPE_CHECKING:
    from pylav.extension.radio.radios import RadioBrowser


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Station:
    radio_api_client: RadioBrowser
    changeuuid: str | None = None
    stationuuid: str | None = None
    serveruuid: str | None = None
    name: str | None = None
    url: str | None = None
    url_resolved: str | None = None
    homepage: str | None = None
    favicon: str | None = None
    tags: str | None = None
    country: str | None = None
    countrycode: str | None = None
    iso_3166_2: str | None = None
    state: str | None = None
    language: str | None = None
    languagecodes: str | None = None
    votes: int | None = None
    lastchangetime: str | None = None
    lastchangetime_iso8601: str | datetime.datetime | None = None
    codec: str | None = None
    bitrate: int | None = None
    hls: int | None = None
    lastcheckok: int | None = None
    lastchecktime: str | None = None
    lastchecktime_iso8601: str | datetime.datetime | None = None
    lastcheckoktime: str | None = None
    lastcheckoktime_iso8601: str | datetime.datetime | None = None
    lastlocalchecktime: str | None = None
    lastlocalchecktime_iso8601: str | datetime.datetime | None = None
    clicktimestamp: str | None = None
    clicktimestamp_iso8601: str | datetime.datetime | None = None
    clickcount: int | None = None
    clicktrend: int | None = None
    ssl_error: int | None = None
    geo_lat: float | None = None
    geo_long: float | None = None
    has_extended_info: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"
        if isinstance(self.lastchangetime_iso8601, str):
            self.lastchangetime_iso8601 = iso8601.parse_date(self.lastchangetime_iso8601)
        if isinstance(self.lastchecktime_iso8601, str):
            self.lastchecktime_iso8601 = iso8601.parse_date(self.lastchecktime_iso8601)
        if isinstance(self.lastcheckoktime_iso8601, str):
            self.lastcheckoktime_iso8601 = iso8601.parse_date(self.lastcheckoktime_iso8601)
        if isinstance(self.lastlocalchecktime_iso8601, str):
            self.lastlocalchecktime_iso8601 = iso8601.parse_date(self.lastlocalchecktime_iso8601)
        if isinstance(self.clicktimestamp_iso8601, str):
            self.clicktimestamp_iso8601 = iso8601.parse_date(self.clicktimestamp_iso8601)

    async def get_query(self) -> Query:
        return await Query.from_string(self.url_resolved or self.url)

    async def click(self) -> None:
        """Increase the click count of a station by one.

        This should be called everytime when a user starts playing a stream to mark the stream more popular than others.
        Every call to this endpoint from the same IP address and for the same station only gets counted once per day.
        """
        await self.radio_api_client.click(station=self)


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Tag:
    name: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Language:
    name: str | None = None
    iso_639: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class State:
    name: str | None = None
    country: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Codec:
    name: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class CountryCode:
    name: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Country:
    name: str | None = None
    iso_3166_1: str | None = None
    stationcount: int | None = None

    def __post_init__(self):
        if self.name is None:
            self.name = "ΩM4L42rPHqy123PyLavInvalidFallback-un5Nht475B"



from __future__ import annotations

import gzip
import io
import sys
from collections.abc import Iterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

import aiohttp
import brotli  # type: ignore
import discord
import ujson
import yaml
from piccolo.columns import Float

from pylav.constants.config import BROTLI_ENABLED
from pylav.constants.playlists import BUNDLED_PLAYLIST_IDS
from pylav.core.context import PyLavContext
from pylav.exceptions.playlist import InvalidPlaylistException
from pylav.logging import getLogger
from pylav.players import filters
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.equalizer import EqualizerRow
from pylav.type_hints.bot import DISCORD_BOT_TYPE

LOGGER = getLogger("PyLav.Database.Equalizer")


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class Equalizer(CachedModel):

    id: int
    scope: int
    author: int
    name: str | None = None
    description: str | None = None
    band_25: float | None = None
    band_40: float | None = None
    band_63: float | None = None
    band_100: float | None = None
    band_160: float | None = None
    band_250: float | None = None
    band_400: float | None = None
    band_630: float | None = None
    band_1000: float | None = None
    band_1600: float | None = None
    band_2500: float | None = None
    band_4000: float | None = None
    band_6300: float | None = None
    band_10000: float | None = None
    band_16000: float | None = None

    def get_cache_key(self) -> str:
        return f"{self.scope}:{self.id}:{self.author}"

    async def save(self) -> Equalizer:
        """Save the Equalizer to the database.

        Returns
        -------
        Equalizer
            The Equalizer.
        """

        values = {
            EqualizerRow.scope: self.scope,
            EqualizerRow.author: self.author,
            EqualizerRow.name: self.name,
            EqualizerRow.description: self.description,
        }
        values |= self._get_save_defaults()

        eq = (
            await EqualizerRow.objects()
            .output(load_json=True)
            .get_or_create(EqualizerRow.id == self.id, defaults=values)
        )
        # noinspection PyProtectedMember
        if not eq._was_created:
            await EqualizerRow.update(values).where(EqualizerRow.id == self.id)
        return Equalizer(**eq.to_dict())

    @classmethod
    async def get(cls, identifier: int) -> Equalizer | None:
        """Get an equalizer from the database.

        Parameters
        ----------
        identifier: int
            The id of the equalizer.

        Returns
        -------
        Equalizer | None
            The equalizer if found, else None.
        """
        equalizer = (
            await EqualizerRow.select().where(EqualizerRow.id == identifier).first().output(load_json=True, nested=True)
        )
        return Equalizer(**equalizer) if equalizer else None

    async def delete(self) -> None:
        """Delete the equalizer from the database"""
        await EqualizerRow.delete().where(EqualizerRow.id == self.id)

    async def can_manage(self, bot: DISCORD_BOT_TYPE, requester: discord.abc.User) -> bool:  # noqa
        """Check if the requester can manage the equalizer.

        Parameters
        ----------
        bot: DISCORD_BOT_TYPE
            The bot.
        requester: discord.abc.User
            The requester.

        Returns
        -------
        bool
            If the requester can manage the equalizer.
        """
        if self.scope in BUNDLED_PLAYLIST_IDS:
            return False
        if requester.id in ((ids := getattr(bot, "owner_ids")) or ()) or requester.id == bot.owner_id:  # noqa
            return True
        elif self.scope == bot.user.id:
            return False
        return self.author == requester.id

    async def get_scope_name(self, bot: DISCORD_BOT_TYPE, mention: bool = True, guild: discord.Guild = None) -> str:
        """Get the name of the scope.

        Parameters
        ----------
        bot: DISCORD_BOT_TYPE
            The bot.
        mention: bool
            If the name should be mentionable.
        guild: discord.Guild | None
            The guild.

        Returns
        -------
        str
            The name of the scope.
        """
        if bot.user.id == self.scope:
            return f"(Global) {bot.user.mention}" if mention else f"(Global) {bot.user}"
        elif guild_ := bot.get_guild(self.scope):
            if guild_:
                guild = guild_
            return f"(Server) {guild.name}"
        elif guild and (channel := guild.get_channel_or_thread(self.scope)):
            return f"(Channel) {channel.mention}" if mention else f"(Channel) {channel.name}"

        elif (
            (guild := guild_ or guild)
            and (guild and (author := guild.get_member(self.scope)))  # noqa
            or (author := bot.get_user(self.author))
        ):
            return f"(User) {author.mention}" if mention else f"(User) {author}"
        else:
            return f"(Invalid) {self.scope}"

    async def get_author_name(self, bot: DISCORD_BOT_TYPE, mention: bool = True) -> str | None:
        """Get the name of the author.

        Parameters
        ----------
        bot: DISCORD_BOT_TYPE
            The bot.
        mention: bool
            If the name should be mentionable.

        Returns
        -------
        str | None
            The name of the author.
        """
        if user := bot.get_user(self.author):
            return f"{user.mention}" if mention else f"{user}"
        return f"{self.author}"

    @asynccontextmanager
    async def to_yaml(self, guild: discord.Guild) -> Iterator[tuple[io.BytesIO, str | None]]:
        """Serialize the Equalizer to a YAML file.

        yields a tuple of (io.BytesIO, bool) where the bool is whether the playlist file was compressed using Gzip

        Parameters
        ----------
        guild: discord.Guild
            The guild.

        Yields
        -------
        tuple[io.BytesIO, str | None]
            The YAML file and the compression type.

        """
        data = {
            "id": self.id,
            "name": self.name,
            "author": self.author,
            "description": self.description,
            "scope": self.scope,
            "bands": self._get_band_dict(),
        }
        compression = None
        with io.BytesIO() as bio:
            yaml.safe_dump(data, bio, default_flow_style=False, sort_keys=False, encoding="utf-8")
            bio.seek(0)
            LOGGER.debug("SIZE UNCOMPRESSED playlist (%s): %s", self.name, sys.getsizeof(bio))
            if sys.getsizeof(bio) > guild.filesize_limit:
                with io.BytesIO() as cbio:
                    if BROTLI_ENABLED:
                        compression = "brotli"
                        cbio.write(brotli.compress(yaml.dump(data, encoding="utf-8")))
                    else:
                        compression = "gzip"
                        with gzip.GzipFile(fileobj=cbio, mode="wb", compresslevel=9) as gzip_file:
                            yaml.safe_dump(data, gzip_file, default_flow_style=False, sort_keys=False, encoding="utf-8")
                    cbio.seek(0)
                    LOGGER.debug("SIZE COMPRESSED playlist [%s] (%s): %s", compression, self.name, sys.getsizeof(cbio))
                    yield cbio, compression
                    return
            yield bio, compression

    @classmethod
    async def from_yaml(cls, context: PyLavContext, scope: int, url: str) -> Equalizer:
        """Deserialize a Equalizer from a YAML file.

        Parameters
        ----------
        context: PyLavContext
            The context.
        scope: int
            The scope.
        url: str
            The URL to the YAML file.

        Returns
        -------
        Equalizer
            The Equalizer.

        """
        try:
            async with aiohttp.ClientSession(auto_decompress=False, json_serialize=ujson.dumps) as session:
                async with session.get(url) as response:
                    data = await response.read()
                    if ".gz.pylav" in url:
                        data = gzip.decompress(data)
                    elif ".br.pylav" in url:
                        data = brotli.decompress(data)
                    data = yaml.safe_load(data)
        except Exception as e:
            raise InvalidPlaylistException(f"Invalid equalizer file - {e}") from e

        return cls(**cls._get_args(identifier=context.message.id, scope=scope, data=data))

    def to_dict(self) -> dict[str, int | str | float | dict[str, float]]:
        """Serialize the Equalizer to a dict.

        Returns
        -------
        dict
            The dict representation of the Equalizer.
        """

        return {
            "id": self.id,
            "name": self.name,
            "author": self.author,
            "description": self.description,
            "scope": self.scope,
            "bands": self._get_band_dict(),
        }

    @classmethod
    def from_dict(cls, data: dict) -> Equalizer:
        """Deserialize a Equalizer from a dict.

        Parameters
        ----------
        data: dict
            The data to use to build the Equalizer

        Returns
        -------
        Equalizer
            The Equalizer
        """
        return cls(
            **cls._get_args(data=data),
        )

    def to_filter(self) -> filters.Equalizer:
        """Serialize the Equalizer to a Filter.

        Returns
        -------
        Equalizer
            The filter representation of the Equalizer
        """

        return filters.Equalizer(
            name=self.name or "CustomEqualizer",
            levels=self._get_band_list(),
        )

    @classmethod
    def from_filter(
        cls, equalizer: filters.Equalizer, context: PyLavContext, scope: int, description: str = None
    ) -> Equalizer:
        """Deserialize a Equalizer from a Filter.

        Parameters
        ----------
        equalizer: Equalizer
            The filter object
        context: PyLavContext
            The Context
        scope: int
            The scope number
        description: str
            The description of the Equalizer

        Returns
        -------
        Equalizer
            The Equalizer built from the Equalizer object
        """
        return Equalizer(
            id=context.message.id,
            scope=scope,
            name=equalizer.name,
            author=context.author.id,
            description=description,
            band_25=equalizer.index[0],
            band_40=equalizer.index[1],
            band_63=equalizer.index[2],
            band_100=equalizer.index[3],
            band_160=equalizer.index[4],
            band_250=equalizer.index[5],
            band_400=equalizer.index[6],
            band_630=equalizer.index[7],
            band_1000=equalizer.index[8],
            band_1600=equalizer.index[9],
            band_2500=equalizer.index[10],
            band_4000=equalizer.index[11],
            band_6300=equalizer.index[12],
            band_10000=equalizer.index[13],
            band_16000=equalizer.index[14],
        )

    def _get_band_dict(self) -> dict[str, float]:
        bands = {}
        null_values = [None, 0.0]
        if self.band_25 not in null_values:
            bands["25"] = self.band_25
        if self.band_40 not in null_values:
            bands["40"] = self.band_40
        if self.band_63 not in null_values:
            bands["63"] = self.band_63
        if self.band_100 not in null_values:
            bands["100"] = self.band_100
        if self.band_160 not in null_values:
            bands["160"] = self.band_160
        if self.band_250 not in null_values:
            bands["250"] = self.band_250
        if self.band_400 not in null_values:
            bands["400"] = self.band_400
        if self.band_630 not in null_values:
            bands["630"] = self.band_630
        if self.band_1000 not in null_values:
            bands["1000"] = self.band_1000
        if self.band_1600 not in null_values:
            bands["1600"] = self.band_1600
        if self.band_2500 not in null_values:
            bands["2500"] = self.band_2500
        if self.band_4000 not in null_values:
            bands["4000"] = self.band_4000
        if self.band_6300 not in null_values:
            bands["6300"] = self.band_6300
        if self.band_10000 not in null_values:
            bands["10000"] = self.band_10000
        if self.band_16000 not in null_values:
            bands["16000"] = self.band_16000
        return bands

    def _get_save_defaults(self) -> dict[Float, float]:
        return {
            EqualizerRow.band_25: self.band_25 or None,
            EqualizerRow.band_40: self.band_40 or None,
            EqualizerRow.band_63: self.band_63 or None,
            EqualizerRow.band_100: self.band_100 or None,
            EqualizerRow.band_160: self.band_160 or None,
            EqualizerRow.band_250: self.band_250 or None,
            EqualizerRow.band_400: self.band_400 or None,
            EqualizerRow.band_630: self.band_630 or None,
            EqualizerRow.band_1000: self.band_1000 or None,
            EqualizerRow.band_1600: self.band_1600 or None,
            EqualizerRow.band_2500: self.band_2500 or None,
            EqualizerRow.band_4000: self.band_4000 or None,
            EqualizerRow.band_6300: self.band_6300 or None,
            EqualizerRow.band_10000: self.band_10000 or None,
            EqualizerRow.band_16000: self.band_16000 or None,
        }

    def _get_band_list(self) -> list[dict[str, float]]:
        levels = []
        null_values = [None, 0.0]
        if self.band_25 not in null_values:
            levels.append({"band": 0, "gain": self.band_25})
        if self.band_40 not in null_values:
            levels.append({"band": 1, "gain": self.band_40})
        if self.band_63 not in null_values:
            levels.append({"band": 2, "gain": self.band_63})
        if self.band_100 not in null_values:
            levels.append({"band": 3, "gain": self.band_100})
        if self.band_160 not in null_values:
            levels.append({"band": 4, "gain": self.band_160})
        if self.band_250 not in null_values:
            levels.append({"band": 5, "gain": self.band_250})
        if self.band_400 not in null_values:
            levels.append({"band": 6, "gain": self.band_400})
        if self.band_630 not in null_values:
            levels.append({"band": 7, "gain": self.band_630})
        if self.band_1000 not in null_values:
            levels.append({"band": 8, "gain": self.band_1000})
        if self.band_1600 not in null_values:
            levels.append({"band": 9, "gain": self.band_1600})
        if self.band_2500 not in null_values:
            levels.append({"band": 10, "gain": self.band_2500})
        if self.band_4000 not in null_values:
            levels.append({"band": 11, "gain": self.band_4000})
        if self.band_6300 not in null_values:
            levels.append({"band": 12, "gain": self.band_6300})
        if self.band_10000 not in null_values:
            levels.append({"band": 13, "gain": self.band_10000})
        if self.band_16000 not in null_values:
            levels.append({"band": 14, "gain": self.band_16000})
        return levels

    @staticmethod
    def _get_args(
        data: dict[str, dict[str, float]],
        identifier: int | None = None,
        scope: int | None = None,
        name: str | None = None,
        author: int | None = None,
        description: str | None = None,
    ) -> dict[str, int | str | float | None]:
        return dict(
            band_25=data["bands"]["25"] if "25" in data["bands"] else None,
            band_40=data["bands"]["40"] if "40" in data["bands"] else None,
            band_63=data["bands"]["63"] if "63" in data["bands"] else None,
            band_100=data["bands"]["100"] if "100" in data["bands"] else None,
            band_160=data["bands"]["160"] if "160" in data["bands"] else None,
            band_250=data["bands"]["250"] if "250" in data["bands"] else None,
            band_400=data["bands"]["400"] if "400" in data["bands"] else None,
            band_630=data["bands"]["630"] if "630" in data["bands"] else None,
            band_1000=data["bands"]["1000"] if "1000" in data["bands"] else None,
            band_1600=data["bands"]["1600"] if "1600" in data["bands"] else None,
            band_2500=data["bands"]["2500"] if "2500" in data["bands"] else None,
            band_4000=data["bands"]["4000"] if "4000" in data["bands"] else None,
            band_6300=data["bands"]["6300"] if "6300" in data["bands"] else None,
            band_10000=data["bands"]["10000"] if "10000" in data["bands"] else None,
            band_16000=data["bands"]["16000"] if "16000" in data["bands"] else None,
            id=identifier or data["id"],
            scope=scope or data["scope"],
            name=name or data["name"],
            author=author or data["author"],
            description=description or data["description"],
        )



from __future__ import annotations

import asyncio

import discord

from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

LOGGER = getLogger("PyLav.ext.red.ui.modals.generic")


class PromptForInput(discord.ui.Modal):
    interaction: DISCORD_INTERACTION_TYPE
    response: str

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        title: str,
        label: str,
        timeout: float | None = None,
        placeholder: str = None,
        style: discord.TextStyle = discord.TextStyle.paragraph,
        min_length: int = 1,
        max_length: int = 64,
        row: int = 1,
    ):
        super().__init__(title=title, timeout=timeout)
        self.cog = cog
        self.text = discord.ui.TextInput(
            label=label, style=style, placeholder=placeholder, min_length=min_length, max_length=max_length, row=row
        )
        self.add_item(self.text)
        self.responded = asyncio.Event()
        self.response = None  # type: ignore
        self.interaction = None  # type: ignore

    async def on_submit(self, interaction: DISCORD_INTERACTION_TYPE):
        self.interaction = interaction
        await interaction.response.defer()
        self.responded.set()
        self.response = self.text.value.strip()



from __future__ import annotations

import dataclasses
from typing import Literal  # noqa

from pylav.nodes.api.responses.player import State
from pylav.nodes.api.responses.rest_api import LavalinkException as TrackExceptionClass


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class CPU:
    cores: int
    systemLoad: float
    lavalinkLoad: float


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Memory:
    free: int
    allocated: int
    reservable: int
    used: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Frame:
    sent: int
    nulled: int
    deficit: int


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Message:
    op: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Ready(Message):
    sessionId: str | None = None
    resumed: bool = False


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Stats(Message):
    players: int = 0
    playingPlayers: int = 0
    uptime: int = 0
    memory: Memory | None | dict = None
    cpu: CPU | None | dict = None
    frameStats: Frame | None | dict = None
    uptime_seconds: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "uptime_seconds", self.uptime / 1000)
        if isinstance(self.memory, dict):
            object.__setattr__(self, "memory", Memory(**self.memory))
        if isinstance(self.cpu, dict):
            object.__setattr__(self, "cpu", CPU(**self.cpu))
        if isinstance(self.frameStats, dict):
            object.__setattr__(self, "frameStats", Frame(**self.frameStats))


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class PlayerUpdate(Message):
    guildId: str | None = None
    state: State | dict | None = None

    def __post_init__(self) -> None:
        if isinstance(self.state, dict):
            object.__setattr__(self, "state", State(**self.state))


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackStart(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["TrackStartEvent"] = "TrackStartEvent"
    encodedTrack: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackStuck(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["TrackStuckEvent"] = "TrackStuckEvent"
    thresholdMs: int | None = None
    encodedTrack: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackEnd(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["TrackEndEvent"] = "TrackEndEvent"
    reason: Literal["FINISHED", "LOAD_FAILED", "STOPPED", "REPLACED", "CLEANUP"] | None = None
    encodedTrack: str | None = None


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class TrackException(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["TrackExceptionEvent"] = "TrackExceptionEvent"
    exception: TrackExceptionClass | dict | None = None
    encodedTrack: str | None = None

    def __post_init__(self) -> None:
        if isinstance(self.exception, dict):
            object.__setattr__(self, "exception", TrackExceptionClass(**self.exception))


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Closed(Message):
    op: Literal["event"] = "event"
    guildId: str | None = None
    type: Literal["WebSocketClosedEvent"] = "WebSocketClosedEvent"
    code: int | None = None
    reason: str | None = None
    byRemote: bool = False



"""
MIT License

Copyright (c) 2017-present Devoxin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import struct
import typing
from base64 import b64decode, b64encode
from io import BytesIO


# noinspection SpellCheckingInspection
class DataReader:
    def __init__(self, ts: str) -> None:
        self._buf = BytesIO(b64decode(ts))
        # Added for PyLav
        self._flag_read = False
        self._flags = 0
        self._version_read = False
        self._version = 0

    def _read(self, count: int) -> bytes:
        return self._buf.read(count)

    def read_byte(self) -> bytes:
        return self._read(1)

    def read_boolean(self) -> bool:
        (result,) = struct.unpack("B", self.read_byte())
        return bool(result)

    def read_unsigned_short(self) -> int:
        (result,) = struct.unpack(">H", self._read(2))
        return typing.cast(int, result)

    def read_int(self) -> int:
        (result,) = struct.unpack(">i", self._read(4))
        return typing.cast(int, result)

    def read_long(self) -> int:
        (result,) = struct.unpack(">Q", self._read(8))
        return typing.cast(int, result)

    def read_utf(self) -> str:
        text_length = self.read_unsigned_short()
        return self._read(text_length).decode()

    def read_utfm(self) -> str:
        text_length = self.read_unsigned_short()
        utf_string = self._read(text_length)
        return self._read_utfm(text_length, utf_string)

    # Merged from utfm_codec.py
    @staticmethod
    def _read_utfm(utf_len: int, utf_bytes: bytes) -> str:
        chars = []
        count = 0

        while count < utf_len:
            c = utf_bytes[count] & 0xFF
            if c > 127:
                break

            count += 1
            chars.append(chr(c))

        while count < utf_len:
            c = utf_bytes[count] & 0xFF
            shift = c >> 4

            if 0 <= shift <= 7:
                count += 1
                chars.append(chr(c))
            elif 12 <= shift <= 13:
                count += 2
                if count > utf_len:
                    raise UnicodeError("Malformed input: partial character at end")
                char2 = utf_bytes[count - 1]
                if (char2 & 0xC0) != 0x80:
                    raise UnicodeError(f"Malformed input around byte {count}")

                char_shift = ((c & 0x1F) << 6) | (char2 & 0x3F)
                chars.append(chr(char_shift))
            elif shift == 14:
                count += 3
                if count > utf_len:
                    raise UnicodeError("Malformed input: partial character at end")

                char2 = utf_bytes[count - 2]
                char3 = utf_bytes[count - 1]

                if (char2 & 0xC0) != 0x80 or (char3 & 0xC0) != 0x80:
                    raise UnicodeError(f"Malformed input around byte {count - 1}")

                char_shift = ((c & 0x0F) << 12) | ((char2 & 0x3F) << 6) | ((char3 & 0x3F) << 0)
                chars.append(chr(char_shift))
            else:
                raise UnicodeError(f"malformed input around byte {count}")

        # noinspection SpellCheckingInspection
        return "".join(chars).encode("utf-16", "surrogatepass").decode("utf-16")

    # Added for PyLav
    def read_nullable_utf(self) -> str | None:
        return self.read_utf() if self.read_boolean() else None

    # Added for PyLav
    def read_nullable_utfm(self) -> str | None:
        return self.read_utfm() if self.read_boolean() else None

    # Added for PyLav
    def read_flags(self) -> int:
        if self._flag_read:
            return self._flags
        self._flags = (self.read_int() & 0xC0000000) >> 30
        self._flag_read = True
        return self._flags

    # Added for PyLav
    def read_version(self) -> int:
        if self._version_read:
            return self._version
        (version,) = (struct.unpack("B", self.read_byte())) if self.read_flags() & 1 != 0 else (1,)
        self._version = version
        self._version_read = True
        return self._version


class DataWriter:
    def __init__(self) -> None:
        self._buf = BytesIO()

    def _write(self, data: bytes) -> None:
        self._buf.write(data)

    def write_byte(self, byte: bytes) -> None:
        self._buf.write(byte)

    def write_boolean(self, boolean: bool) -> None:
        enc = struct.pack("B", 1 if boolean else 0)
        self.write_byte(enc)

    def write_unsigned_short(self, short: int) -> None:
        enc = struct.pack(">H", short)
        self._write(enc)

    def write_int(self, integer: int) -> None:
        enc = struct.pack(">i", integer)
        self._write(enc)

    def write_long(self, long_value: int) -> None:
        enc = struct.pack(">Q", long_value)
        self._write(enc)

    def write_utf(self, utf_string: str) -> None:
        utf = utf_string.encode("utf8")
        byte_len = len(utf)

        if byte_len > 65535:
            raise OverflowError("UTF string may not exceed 65535 bytes!")

        self.write_unsigned_short(byte_len)
        self._write(utf)

    def finish(self) -> bytes:
        with BytesIO() as track_buf:
            # Simplified for PyLav
            track_buf.write(self.get_flags())
            self._buf.seek(0)
            track_buf.write(self._buf.read())
            self._buf.close()
            track_buf.seek(0)
            return track_buf.read()

    # Added for PyLav
    def write_nullable_utf(self, utf_string: str | None) -> None:
        if utf_string is None:
            self.write_boolean(False)
        else:
            self.write_boolean(True)
            self.write_utf(utf_string)

    # Added for PyLav
    def write_version(self, version: int) -> None:
        self.write_byte(struct.pack("B", version))

    # Added for PyLav
    def get_flags(self) -> bytes:
        byte_len = self._buf.getbuffer().nbytes
        flags = byte_len | (1 << 30)
        return struct.pack(">i", flags)

    # Added for PyLav
    def to_base64(self) -> str:
        return b64encode(self.finish()).decode()



from __future__ import annotations

import threading

from asyncpg import UniqueViolationError  # type: ignore
from piccolo.columns import M2M, LazyTableReference, Text
from piccolo.columns.indexes import IndexMethod
from piccolo.table import Table

from pylav.helpers.singleton import synchronized_method_call
from pylav.logging import getLogger
from pylav.storage.database.tables.misc import DATABASE_ENGINE
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

LOGGER = getLogger("PyLav.Database.Track")
_LOCK = threading.Lock()


class TrackRow(Table, db=DATABASE_ENGINE, tablename="track"):
    identifier = Text(null=True, default=None, index=True)
    sourceName = Text(null=True, default=None, index=True)
    title = Text(null=True, default=None, index=True, index_method=IndexMethod.gin)
    uri = Text(null=True, default=None, index=True)
    # noinspection SpellCheckingInspection
    isrc = Text(null=True, default=None, index=True)
    encoded = Text(null=False, index=True, primary_key=True)
    queries = M2M(LazyTableReference("TrackToQueries", module_path="pylav.storage.database.tables.m2m"))
    playlists = M2M(LazyTableReference("TrackToPlaylists", module_path="pylav.storage.database.tables.m2m"))

    @classmethod
    @synchronized_method_call(_LOCK)
    async def get_or_create(cls, encoded: str, kwargs: JSON_DICT_TYPE) -> TrackRow:
        try:
            return await cls.objects().get_or_create(cls.encoded == encoded, kwargs)
        except UniqueViolationError:
            obj = cls(encoded=encoded, **kwargs)
            obj._exists_in_db = True
            return obj
        except Exception as e:
            LOGGER.trace("Error while creating track: %s", e, exc_info=True)
            raise e



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from discord.app_commands import Choice, Transformer
from discord.ext import commands

from pylav.players.query.obj import Query
from pylav.type_hints.bot import DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    QueryConverter = Query
    QueryPlaylistConverter = Query
else:

    class QueryConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> Query:  # noqa

            arg = arg.strip("<>")
            return await Query.from_string(arg)

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> Query:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        async def autocomplete(self, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            return []

    class QueryPlaylistConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> Query:  # noqa

            arg = arg.strip("<>")
            query = await Query.from_string(arg)
            if not (query.is_playlist or query.is_album):
                raise commands.BadArgument(_("Query must be a playlist or album"))
            return query

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> Query:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        async def autocomplete(self, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            return []



from __future__ import annotations

from dataclasses import _FIELD, _FIELDS, InitVar, asdict, dataclass, field  # noqa
from typing import NotRequired  # noqa

_JSON_TYPE = dict[str, "_JSON_TYPE"] | list["_JSON_TYPE"] | str | int | float | bool | None
_JSON_DICT_TYPE = dict[str, _JSON_TYPE]


@dataclass(kw_only=True)
class PluginInfo:
    kwargs: InitVar[_JSON_DICT_TYPE | None] = None

    def __post_init__(self, kwargs: _JSON_DICT_TYPE | None) -> None:
        if kwargs:
            for k, v in kwargs.items():
                object.__setattr__(self, k, v)

    def __getattr__(self, item):
        return None

    def __repr__(self):
        return f"PluginInfo({', '.join([f'{key}={val!r}' for key, val in self.__dict__.items()])})"

    def to_dict(self) -> _JSON_DICT_TYPE:
        return asdict(self)

    def _add_to_asdict(self, attr: str) -> None:
        f = field(repr=True)
        f.name = attr
        f._field_type = _FIELD
        getattr(self, _FIELDS)[attr] = f


@dataclass(kw_only=True)
class PlaylistPluginInfo(PluginInfo):
    type: NotRequired[str] = field(init=False)
    identifier: NotRequired[str] = field(init=False)
    artworkURL: NotRequired[str] = field(init=False)
    author: NotRequired[str] = field(init=False)


@dataclass(kw_only=True)
class TrackPluginInfo(PluginInfo):
    probeInfo: NotRequired[str] = field(init=False)



from __future__ import annotations

import asyncio
import contextlib
import pathlib
from collections.abc import Iterator
from typing import TYPE_CHECKING

import asyncpg
import discord

from pylav.events.player import PlayerConnectedEvent
from pylav.exceptions.node import NoNodeAvailableException
from pylav.helpers.format.strings import shorten_string
from pylav.logging import getLogger
from pylav.nodes.node import Node
from pylav.players.player import Player
from pylav.players.query.obj import Query
from pylav.storage.models.player.config import PlayerConfig
from pylav.storage.models.player.state import PlayerState

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


if TYPE_CHECKING:
    from pylav.core.client import Client

LOGGER = getLogger("PyLav.PlayerManager")


class PlayerController:
    """Represents the player manager that contains all the players.

    len(x):
        Returns the total amount of cached players.
    iter(x):
        Returns an iterator of all the players cached.

    Attributes
    ----------
    default_player_class: :class:`BasePlayer`
        The player that the player manager is initialized with.
    bot: :class:`discord.Client`
        The client that the player manager is initialized with.
    client: :class:`Client`
        The client that the player manager is initialized with.
    """

    __slots__ = ("_players", "default_player_class", "bot", "client", "_global_player_config")

    _global_player_config: PlayerConfig

    def __init__(self, lavalink: Client, player: type[Player] = Player):
        if not issubclass(player, Player):
            raise ValueError("Player must implement Player")

        self.client = lavalink
        self.bot = lavalink.bot
        self._players: dict[int, Player] = {}
        self.default_player_class = player

    def __len__(self):
        return len(self._players)

    def __iter__(self) -> Iterator[tuple[int, Player]]:
        """Returns an iterator that yields a tuple of (guild_id, player)"""
        yield from self.players.items()

    @property
    def players(self) -> dict[int, Player]:
        """Returns a dictionary of all players in manager."""
        return self._players

    @property
    def global_config(self) -> PlayerConfig:
        return self._global_player_config

    @property
    def connected_players(self) -> list[Player]:
        """Returns a list of all the connected players"""
        return [p for p in self.players.values() if p.is_connected]

    @property
    def playing_players(self) -> list[Player]:
        """Returns a list of all the playing players"""
        return [p for p in self.players.values() if p.is_playing]

    @property
    def not_playing_players(self) -> list[Player]:
        """Returns a list of all the not playing players"""
        return [p for p in self.players.values() if not p.is_playing]

    @property
    def paused_players(self) -> list[Player]:
        """Returns a list of all the paused players"""
        return [p for p in self.players.values() if p.paused]

    @property
    def empty_players(self) -> list[Player]:
        """Returns a list of all the empty players"""
        return [p for p in self.players.values() if p.is_empty]

    async def initialize(self):
        self._global_player_config = self.client.player_config_manager.get_global_config()
        self.client.scheduler.add_job(
            self.update_bot_activity,
            trigger="interval",
            seconds=5,
            max_instances=1,
            replace_existing=True,
            name="update_bot_activity",
            coalesce=True,
            id=f"{self.bot.user.id}-update_bot_activity",
        )

    async def destroy(self, guild_id: int, requester: discord.Member | None):
        """
        Removes a player from cache, and also Lavalink if applicable.
        Ensure you have disconnected the given guild_id from the voicechannel
        first, if connected.
        Warning
        -------
        This should only be used if you know what you're doing. Players should never be
        destroyed unless they have been moved to another :class:`Node`.
        Parameters
        ----------
        guild_id: int
            The guild_id associated with the player to remove.
        requester: :class:`discord.Member`
            The member requesting the player to be removed.
        """
        if guild_id not in self.players:
            return

        player = self.players.pop(guild_id)

        if player.node and player.node.available:
            await player.node.delete_session_player(player.guild.id)
        await player.disconnect(requester=requester)
        # noinspection PyProtectedMember
        player.node._logger.debug("Successfully destroyed player %s", guild_id)

    async def save_and_restore(self, guild_id: int):
        await asyncio.sleep(5)
        if player := self.players.pop(guild_id, None):
            await player.save()
            await player.disconnect(requester=self.client.bot.user)
        player_state = await self.client.player_state_db_manager.fetch_player(guild_id)
        if player_state:
            await self._restore_player(player_state)

    def find_all(self, predicate=None):
        """Returns a list of players that match the given predicate.

        Parameters
        ----------
        predicate: Optional[:class:`function`]
            A predicate to return specific players. Defaults to `None`.
        Returns
        -------
        List[:class:`Player`]
        """
        return [p for p in self.players.values() if bool(predicate(p))] if predicate else list(self.players.values())

    async def remove(self, guild_id: int) -> None:
        """Removes a player from the internal cache.

        Parameters
        ----------
        guild_id: :class:`int`
            The player that will be removed.
        """
        if guild_id in self.players:
            player = self.players.pop(guild_id)
            player.cleanup()

    def get(self, guild_id: int) -> Player:
        """
        Gets a player from cache.
        Parameters
        ----------
        guild_id: :class:`int`
            The guild_id associated with the player to get.
        Returns
        -------
        Optional[:class:`Player`]
        """
        return self.players.get(guild_id)

    async def create(
        self,
        channel: discord.channel.VocalGuildChannel,
        endpoint: str = None,
        node: Node = None,
        self_deaf: bool = None,
        requester: discord.Member = None,
        feature: str | None = None,
    ) -> Player:
        """
        Creates a player if one doesn't exist with the given information.
        If node is provided, a player will be created on that node.
        If endpoint is provided, PyLav will attempt to parse the region from the endpoint
        and return a node in the parsed region.
        If node, region and endpoint are left unspecified, or region/endpoint selection fails,
        PyLav will fall back to the node with the lowest penalty.
        Region can be omitted if node is specified and vice-versa.
        Parameters
        ----------
        channel: :class:`discord.channel.VocalGuildChannel`
            The voice channel to connect to.
        endpoint: :class:`str`
            The address of the Discord voice server. Defaults to `None`.
        node: :class:`Node`
            The node to put the player on. Defaults to `None` and a node with the lowest penalty is chosen.
        requester: :class:`discord.Member`
            The member requesting the player. Defaults to `None`.
        feature: Optional[:class:`str`]
            The feature to look for for the initial Node. Defaults to `None`.
        self_deaf: :class:`bool`
            Whether the player should deafen themselves. Defaults to `False`.
        Returns
        -------
        :class:`Player`
        """
        if p := self.players.get(channel.guild.id):
            if channel.id != p.channel_id:
                await p.move_to(requester, channel)
            return p

        region = self.client.node_manager.get_region(endpoint)

        best_node = node or await self.client.node_manager.find_best_node(region, feature=feature or None)
        if not best_node:
            raise NoNodeAvailableException("No available nodes!")
        player_config = self.client.player_config_manager.get_config(channel.guild.id)
        forced_channel_id = await player_config.fetch_forced_channel_id()
        self_deafen = await self.client.player_config_manager.get_self_deaf(channel.guild.id)
        if forced_channel_id != 0:
            act_channel = channel.guild.get_channel_or_thread(forced_channel_id)
        else:
            act_channel = channel
        player: Player = await act_channel.connect(
            cls=Player, self_deaf=self_deafen if self_deaf is None else self_deaf  # type: ignore
        )
        self.players[channel.guild.id] = player
        try:
            best_node = node or await self.client.node_manager.find_best_node(
                region, feature=feature or None, coordinates=player.coordinates
            )
            if not best_node:
                raise NoNodeAvailableException(_("No available nodes!"))
            await player.post_init(
                node=best_node, player_manager=self, config=player_config, pylav=self.client, requester=requester
            )
            await player.move_to(
                requester, channel=player.channel, self_deaf=self_deafen if self_deaf is None else self_deaf
            )
            best_node.dispatch_event(PlayerConnectedEvent(player, requester or self.client.bot.user))
            # noinspection PyProtectedMember
            best_node._logger.debug("Successfully created player for %s", channel.guild.id)
            return player
        except Exception:
            LOGGER.error("Failed to create player for %s", channel.guild.id)
            LOGGER.debug("Error in create player for %s", channel.guild.id, exc_info=True)
            await player.disconnect(requester=requester)
            raise

    async def save_all_players(self) -> None:
        LOGGER.debug("Saving player states")
        await self.client.player_state_db_manager.save_players([await p.to_dict() for p in self.connected_players])

    async def restore_player_states(self) -> None:
        # noinspection PyProtectedMember
        await asyncio.wait_for(self.client._wait_for_playlists.wait(), timeout=600)
        LOGGER.info("Restoring player states")
        while not self.client.node_manager.available_nodes:
            await asyncio.sleep(1)
        tasks = [
            asyncio.create_task(self._restore_player(p))
            async for p in self.client.player_state_db_manager.fetch_all_players()
        ]
        await asyncio.gather(*tasks, return_exceptions=True)
        LOGGER.info("Restored %s player states", len(self.players))

    async def _restore_player(self, player_state: PlayerState) -> None:

        player = self.players.get(player_state.id)
        if player is not None:
            # Player was started before restore
            LOGGER.debug("Player %s initialized before restore, skipping restore", player_state.id)
            await self.client.player_state_db_manager.delete_player(guild_id=player_state.id)
            return
        channel = self.client.bot.get_channel(player_state.channel_id)
        if not channel:
            # Channel does not exist anymore
            LOGGER.debug("Channel for %s could not be found, skipping player restore", player_state.id)
            await self.client.player_state_db_manager.delete_player(guild_id=player_state.id)
            return
        requester = self.client.bot.user
        try:
            discord_player = await self.create(
                channel=channel,
                requester=requester,
                feature=(await Query.from_base64(player_state.current["encoded"], lazy=True)).requires_capability
                if player_state.current
                else None,
                self_deaf=player_state.self_deaf,
            )
        except Exception:
            LOGGER.exception("Failed to restore player %s - %s", player_state.id, player_state.channel_id)
            raise
        # noinspection PyProtectedMember
        if not discord_player._restored:
            await discord_player.restore(player_state, requester)

    async def shutdown(self) -> None:
        LOGGER.info("Shutting down all players")
        tasks = [
            asyncio.create_task(self.destroy(guild_id=guild_id, requester=self.client.bot.user))
            for guild_id in self.players
        ]
        await asyncio.gather(*tasks, return_exceptions=True)

    async def update_bot_activity(self) -> None:
        """
        Updates the bot's activity.
        """
        with contextlib.suppress(asyncio.exceptions.CancelledError, asyncpg.exceptions.CannotConnectNowError):
            if not await (self.client.lib_db_manager.get_config()).fetch_update_bot_activity():
                return
            playing_players = len(self.playing_players)
            activities = self.bot.guilds[0].me.activities
            activity = discord.utils.find(lambda a: a.type == discord.ActivityType.listening, activities)
            if playing_players > 1:
                if (not activity) or f"Music in {playing_players} servers" not in activity.name:
                    LOGGER.debug("Updating bot activity to %s", f"Listening to Music in {playing_players} servers")
                    await self.bot.change_presence(
                        activity=discord.Activity(
                            type=discord.ActivityType.listening,
                            name=shorten_string(
                                max_length=100, string=_("Music in {count} servers").format(count=playing_players)
                            ),
                        )
                    )
            elif playing_players == 1:
                current_player = self.playing_players[0]
                if current_player.current is None:
                    return
                track_name = await current_player.current.get_track_display_name(
                    max_length=40,
                    author=True,
                    unformatted=True,
                )
                if activity and track_name in activity.name:
                    return
                LOGGER.debug("Updating bot activity to %s", f"Listening to {track_name}")
                await self.bot.change_presence(
                    activity=discord.Activity(type=discord.ActivityType.listening, name=track_name)
                )

            elif playing_players == 0 and activity:
                LOGGER.debug("Removing bot activity")
                await self.bot.change_presence(activity=None)



from __future__ import annotations

import dataclasses


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Distortion:
    sinOffset: float | None = None
    sinScale: float | None = None
    cosOffset: float | None = None
    cosScale: float | None = None
    tanOffset: float | None = None
    tanScale: float | None = None
    offset: float | None = None
    scale: float | None = None

    def to_dict(self) -> dict[str, float]:
        response = {}
        if self.sinOffset is not None:
            response["sinOffset"] = self.sinOffset
        if self.sinScale is not None:
            response["sinScale"] = self.sinScale
        if self.cosOffset is not None:
            response["cosOffset"] = self.cosOffset
        if self.cosScale is not None:
            response["cosScale"] = self.cosScale
        if self.tanOffset is not None:
            response["tanOffset"] = self.tanOffset
        if self.tanScale is not None:
            response["tanScale"] = self.tanScale
        if self.offset is not None:
            response["offset"] = self.offset
        if self.scale is not None:
            response["scale"] = self.scale
        return response



from __future__ import annotations

import datetime
import pathlib
import typing
from dataclasses import dataclass

import aiopath  # type: ignore
import ujson

from pylav.extension.bundled_node.utils import get_true_path
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.helpers.time import get_now_utc
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.config import LibConfigRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE, JSON_DICT_WITH_DATE_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Config(CachedModel, metaclass=SingletonCachedByKey):
    bot: int
    id: int = 1

    def get_cache_key(self) -> str:
        return f"{self.id}:{self.bot}:"

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the config exists.

        Returns
        -------
        bool
            Whether the config exists.
        """
        return await LibConfigRow.exists().where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))

    @maybe_cached
    async def fetch_config_folder(self) -> str:
        """Fetch the config folder.

        Returns
        -------
        str
            The config folder.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.config_folder)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["config_folder"] if response else LibConfigRow.config_folder.default

    async def update_config_folder(self, config_folder: aiopath.AsyncPath | pathlib.Path | str) -> None:
        """Update the config folder.

        Parameters
        ----------
        config_folder
            The new config folder.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, config_folder)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET config_folder = EXCLUDED.config_folder
            """,
            self.id,
            self.bot,
            str(config_folder),
        )
        await self.update_cache((self.fetch_config_folder, str(config_folder)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_localtrack_folder(self) -> str:
        """Fetch the localtrack folder.

        Returns
        -------
        str
            The localtrack folder.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.localtrack_folder)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["localtrack_folder"] if response else LibConfigRow.localtrack_folder.default

    async def update_localtrack_folder(self, localtrack_folder: aiopath.AsyncPath | pathlib.Path | str) -> None:
        """Update the localtrack folder.

        Parameters
        ----------
        localtrack_folder
            The new localtrack folder.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, localtrack_folder)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET localtrack_folder = EXCLUDED.localtrack_folder
            """,
            self.id,
            self.bot,
            str(localtrack_folder),
        )
        await self.update_cache((self.fetch_localtrack_folder, str(localtrack_folder)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_java_path(self) -> str:
        """Fetch the java path.

        Returns
        -------
        str
            The java path.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.java_path)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )

        temp_path = response["java_path"] if response else LibConfigRow.java_path.default
        java_path = get_true_path(temp_path, temp_path)
        return java_path

    async def update_java_path(self, java_path: aiopath.AsyncPath | pathlib.Path | str) -> None:
        """Update the java path.

        Parameters
        ----------
        java_path
            The new java path.
        """
        java_path = get_true_path(java_path, java_path)
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, java_path)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET java_path = EXCLUDED.java_path
            """,
            self.id,
            self.bot,
            str(java_path),
        )
        await self.update_cache((self.fetch_java_path, str(java_path)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_enable_managed_node(self) -> bool:
        """Fetch the enable_managed_node.

        Returns
        -------
        bool
            The enable_managed_node.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.enable_managed_node)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["enable_managed_node"] if response else LibConfigRow.enable_managed_node.default

    async def update_enable_managed_node(self, enable_managed_node: bool) -> None:
        """Update the enable_managed_node.

        Parameters
        ----------
        enable_managed_node
            The new enable_managed_node.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, enable_managed_node)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET enable_managed_node = EXCLUDED.enable_managed_node
            """,
            self.id,
            self.bot,
            enable_managed_node,
        )
        await self.update_cache((self.fetch_enable_managed_node, enable_managed_node), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_use_bundled_pylav_external(self) -> bool:
        """Fetch the use_bundled_pylav_external.

        Returns
        -------
        bool
            The use_bundled_pylav_external.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.use_bundled_pylav_external)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["use_bundled_pylav_external"] if response else LibConfigRow.use_bundled_pylav_external.default

    async def update_use_bundled_pylav_external(self, use_bundled_pylav_external: bool) -> None:
        """Update the use_bundled_pylav_external.

        Parameters
        ----------
        use_bundled_pylav_external
            The new use_bundled_pylav_external.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, use_bundled_pylav_external)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET use_bundled_pylav_external = EXCLUDED.use_bundled_pylav_external
            """,
            self.id,
            self.bot,
            use_bundled_pylav_external,
        )
        await self.update_cache(
            (self.fetch_use_bundled_pylav_external, use_bundled_pylav_external), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_use_bundled_lava_link_external(self) -> bool:
        """Fetch the use_bundled_lava_link_external.

        Returns
        -------
        bool
            The use_bundled_lava_link_external.
        """
        return False

    async def update_use_bundled_lava_link_external(self, use_bundled_lava_link_external: bool) -> None:
        """Update the use_bundled_lava_link_external.

        Parameters
        ----------
        use_bundled_lava_link_external
            The new use_bundled_lava_link_external.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, use_bundled_lava_link_external)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET use_bundled_lava_link_external = EXCLUDED.use_bundled_lava_link_external
            """,
            self.id,
            self.bot,
            False,
        )
        await self.update_cache((self.fetch_use_bundled_lava_link_external, False), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_download_id(self) -> int:
        """Fetch the download_id.

        Returns
        -------
        str
            The download_id.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.download_id)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["download_id"] if response else LibConfigRow.download_id.default

    async def update_download_id(self, download_id: int) -> None:
        """Update the download_id.

        Parameters
        ----------
        download_id
            The new download_id.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, download_id)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET download_id = EXCLUDED.download_id
            """,
            self.id,
            self.bot,
            download_id,
        )
        await self.update_cache((self.fetch_download_id, download_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the extras.

        Returns
        -------
        dict
            The extras.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.extras)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["extras"] if response else ujson.loads(LibConfigRow.extras.default)

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the extras.

        Parameters
        ----------
        extras
            The new extras.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, extras)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET extras = EXCLUDED.extras
            """,
            self.id,
            self.bot,
            ujson.dumps(extras),
        )
        await self.update_cache((self.fetch_extras, extras), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_next_execution_update_bundled_playlists(self) -> datetime.datetime:
        """Fetch the next_execution_update_bundled_playlists.

        Returns
        -------
        datetime.datetime
            The next_execution_update_bundled_playlists.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.next_execution_update_bundled_playlists)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["next_execution_update_bundled_playlists"] if response else get_now_utc()

    async def update_next_execution_update_bundled_playlists(self, next_execution: datetime.datetime) -> None:
        """Update the next_execution_update_bundled_playlists.

        Parameters
        ----------
        next_execution
            The new next_execution_update_bundled_playlists.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, next_execution_update_bundled_playlists)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET next_execution_update_bundled_playlists = EXCLUDED.next_execution_update_bundled_playlists
            """,
            self.id,
            self.bot,
            next_execution,
        )
        await self.update_cache(
            (self.fetch_next_execution_update_bundled_playlists, next_execution), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_next_execution_update_bundled_external_playlists(self) -> datetime.datetime:
        """Fetch the next_execution_update_bundled_external_playlists.

        Returns
        -------
        datetime.datetime
            The next_execution_update_bundled_external_playlists.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.next_execution_update_bundled_external_playlists)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["next_execution_update_bundled_external_playlists"] if response else get_now_utc()

    async def update_next_execution_update_bundled_external_playlists(self, next_execution: datetime.datetime) -> None:
        """Update the next_execution_update_bundled_external_playlists.

        Parameters
        ----------
        next_execution
            The new next_execution.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252

        # noinspection PyPep8
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, next_execution_update_bundled_external_playlists)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE
            SET next_execution_update_bundled_external_playlists = EXCLUDED.next_execution_update_bundled_external_playlists
            """,  # noqa
            self.id,
            self.bot,
            next_execution,
        )
        await self.update_cache(
            (self.fetch_next_execution_update_bundled_external_playlists, next_execution), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_next_execution_update_external_playlists(self) -> datetime.datetime:
        """Fetch the next_execution_update_external_playlists.

        Returns
        -------
        datetime.datetime
            The next_execution_update_external_playlists.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.next_execution_update_external_playlists)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["next_execution_update_external_playlists"] if response else get_now_utc()

    async def update_next_execution_update_external_playlists(self, next_execution: datetime.datetime) -> None:
        """Update the next_execution_update_external_playlists.

        Parameters
        ----------
        next_execution
            The new next_execution.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, next_execution_update_external_playlists)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET next_execution_update_external_playlists = EXCLUDED.next_execution_update_external_playlists
            """,
            self.id,
            self.bot,
            next_execution,
        )
        await self.update_cache(
            (self.fetch_next_execution_update_external_playlists, next_execution), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_update_bot_activity(self) -> bool:
        """Fetch the update_bot_activity.

        Returns
        -------
        bool
            The update_bot_activity.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.update_bot_activity)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["update_bot_activity"] if response else LibConfigRow.update_bot_activity.default

    async def update_update_bot_activity(self, update_bot_activity: bool) -> None:
        """Update the update_bot_activity.

        Parameters
        ----------
        update_bot_activity
            The new update_bot_activity.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, update_bot_activity)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET update_bot_activity = EXCLUDED.update_bot_activity
            """,
            self.id,
            self.bot,
            update_bot_activity,
        )
        await self.update_cache((self.fetch_update_bot_activity, update_bot_activity), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_update_managed_nodes(self) -> bool:
        """Fetch the auto_update_managed_nodes.

        Returns
        -------
        bool
            The auto_update_managed_nodes.
        """
        response = (
            await LibConfigRow.select(LibConfigRow.auto_update_managed_nodes)
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return response["auto_update_managed_nodes"] if response else LibConfigRow.auto_update_managed_nodes.default

    async def update_auto_update_managed_nodes(self, auto_update_managed_nodes: bool) -> None:
        """Update the auto_update_managed_nodes.

        Parameters
        ----------
        auto_update_managed_nodes
            The new auto_update_managed_nodes.
        """
        # TODO: When piccolo add support to on conflict clauses using RAW here is more efficient
        #  Tracking issue: https://github.com/piccolo-orm/piccolo/issues/252
        await LibConfigRow.raw(
            """
            INSERT INTO lib_config (id, bot, auto_update_managed_nodes)
            VALUES ({}, {}, {})
            ON CONFLICT (id, bot)
            DO UPDATE SET auto_update_managed_nodes = EXCLUDED.auto_update_managed_nodes
            """,
            self.id,
            self.bot,
            auto_update_managed_nodes,
        )
        await self.update_cache((self.fetch_auto_update_managed_nodes, auto_update_managed_nodes), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    async def delete(self) -> None:
        """Delete the config from the database"""
        await LibConfigRow.delete().where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
        await self.invalidate_cache()

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_WITH_DATE_TYPE:
        """Update all attributed for the config from the database.

        Returns
        -------
        Config
            The updated config.
        """
        data = typing.cast(
            JSON_DICT_TYPE,
            await LibConfigRow.select()
            .where((LibConfigRow.id == self.id) & (LibConfigRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True),
        )
        if data:
            data["java_path"] = get_true_path(data["java_path"], data["java_path"])
            return data

        return {
            "id": self.id,
            "bot": self.bot,
            "config_folder": LibConfigRow.config_folder.default,
            "java_path": get_true_path(LibConfigRow.java_path.default),
            "enable_managed_node": LibConfigRow.enable_managed_node.default,
            "auto_update_managed_nodes": LibConfigRow.auto_update_managed_nodes.default,
            "localtrack_folder": LibConfigRow.localtrack_folder.default,
            "download_id": LibConfigRow.download_id.default,
            "update_bot_activity": LibConfigRow.update_bot_activity.default,
            "use_bundled_pylav_external": LibConfigRow.use_bundled_pylav_external.default,
            "use_bundled_lava_link_external": False,
            "extras": ujson.loads(LibConfigRow.extras.default),
            "next_execution_update_bundled_playlists": LibConfigRow.next_execution_update_bundled_playlists.default,
            "next_execution_update_bundled_external_playlists": LibConfigRow.next_execution_update_bundled_external_playlists.default,  # noqa
            "next_execution_update_external_playlists": LibConfigRow.next_execution_update_external_playlists.default,
        }
